#!/usr/bin/env python3

## test distance resolution limit on randomly generated pts or pts from file

import sys
import os
import math
import numpy as np
from scipy import sparse
from scipy.spatial.distance import pdist, squareform
from math import sqrt
from PIL import Image
import glob

myZero = 1e-10
myInfty = 1e30
Density = 0.1
Verbose = False
#Verbose = True
image_exts = [".jpg",".gif",".png"]
thumbnailsize = (100,100)
thumbnaildepth = 3
dtypes = ['U01', 'U', 'N', 'Ns', 'A']

#################### FUNCTIONS #####################

# generate a componentwise Normal(0,1) matrix
def normalmatrix(m, n):
    return np.random.normal([0],[[1]],(m,n))

# generate a componentwise Uniform(0,1) matrix
def uniformmatrix(m, n):
    return np.random.rand(m,n)

# generate a componentwise Normal(0,1) matrix with density s
def sparsenormalmatrix(m,n,s):
    A = np.zeros((m,n))
    for i in range(m):
        for j in range(n):
            if np.random.uniform(0,1) <= s:
                A[i,j] = np.random.randn()
    return A

### generate a sample from an Achlioptas distribution
def achlioptas_sample(p=1.0/3.0):
    ret = 0
    uniformsample = np.random.uniform(0,1)
    if uniformsample < p/2:
        ret = -1
    elif uniformsample > 1-p/2:
        ret = 1
    else:
        ret = 0
    return ret
    
### generate an Achlioptas m x n random projection matrix
def achlioptas(m, n, p=1.0/3.0):
    A = np.zeros((m,n))
    for i in range(m):
        for j in range(n):
            aij = achlioptas_sample(p)
            if aij != 0:
                A[i,j] = aij
    return A

####################### MAIN #######################

## read command line
if len(sys.argv) < 2:
    print("syntax: ./distres.py d [K n]")
    print("  d in", dtypes)
    print("    U01 is Uniform(0,1)")
    print("    U is Uniform(-1,1)")
    print("    N is Normal(0,1)")
    print("    Ns is Normal(0,1) with density 0.1")
    print("    A is Achlioptas")
    print("    or d can be a directory containing images")
    print("  K=|dimensions|, n=|points| if d is a distribution")
    print("  default K = 100, default n = 100")
    sys.exit(1)

## generate random data matrix
d = sys.argv[1]
K = 100
n = 100
rndgen = True
if d not in dtypes:
    rndgen = False
if len(sys.argv) >= 3:    
    K = int(sys.argv[2])
if len(sys.argv) >= 4:
    n = int(sys.argv[3])

if d == "U":
    X = uniformmatrix(K,n).T
elif d == "U01":
    X = (uniformmatrix(K,n) + np.ones((K,n))) / 2.0
    X = X.T
elif d == "N":
    X = normalmatrix(K,n).T
elif d == "Ns":
    X = sparsenormalmatrix(K,n,Density).T
elif d == "A":
    X = achlioptas(K,n).T
else:
    rndgen = False
    ## read all image files in dir, scale them, put them into a data matrix
    thsz = thumbnailsize[0]*thumbnailsize[1]*thumbnaildepth
    images = []
    filenames = []
    X = []
    n = 0
    print(sys.argv[0] + ": reading " + d + " ... ", end='')
    sys.stdout.flush()
    for ext in image_exts:
        for filename in glob.glob(d + '/*' + ext):
            im = Image.open(filename)
            im = im.resize(thumbnailsize)
            im = im.convert("RGB")
            images.append(im)
            filenames.append(os.path.basename(filename))
            ima = np.reshape(np.array(im), (thsz))
            X.append(ima)
            n += 1
    X = np.array(X)
    print("done")

sys.stdout.flush()
print("distance resolution limit for ", end='')
if rndgen:
    print("{0:d} pts in {1:d} dim,".format(n,K))
    print("distr {2:s}".format(n,K,d))
else:
    print("images from", d)
    
## compute distance matrix
Dvec = pdist(X)
D = squareform(Dvec)
nD = len(D)

## compute min, max, avg, stdev in one loop per pt
Dmin = {i:myInfty for i in range(nD)}
Dmax = {i:0 for i in range(nD)}
Davg = {i:0 for i in range(nD)}
Dvar = {i:0 for i in range(nD)}
Dstdev = {i:0 for i in range(nD)}
Dreslim = {i:0 for i in range(nD)}
for i in range(nD):
    for j in range(nD):
        if i != j:
            if D[i,j] < Dmin[i]:
                Dmin[i] = D[i,j]
            if D[i,j] > Dmax[i]:
                Dmax[i] = D[i,j]
            # by Knuth according to
            # www.strchr.com/standard_deviation_in_one_pass?allcomments=1
            if j == i+1:
                Davg[i] = D[i,j]
            else:
                delta = float(D[i,j] - Davg[i])
                Davg[i] += delta / nD
                Dvar[i] += delta*(D[i,j] - Davg[i])
    Dvar[i] /= float(nD-1)
    Dstdev[i] = math.sqrt(Dvar[i])
    Dreslim[i] = (Dmax[i] / Dmin[i]) - 1
    if Verbose:
        # print out statistics
        print("point {0:d}".format(i))
        print("  distance res eps = {0:.3g}".format(Dreslim[i]))
        print("  distance mean    = {0:.3g}".format(Davg[i]))
        print("  distance stdev   = {0:.3g}".format(Dstdev[i]))

print("avg distance res limit epsilon = {0:.5g}".format(np.nanmean(np.array(list(Dreslim.values())))))
print("stdev on dst res limit epsilon = {0:.5g}".format(np.nanstd(np.array(list(Dreslim.values())))))
print("avg mean distance  = {0:.5g}".format(np.nanmean(np.array(list(Davg.values())))))
print("min min distance   = {0:.5g}".format(np.nanmin(np.array(list(Dmin.values())))))
print("max max distance   = {0:.5g}".format(np.nanmax(np.array(list(Dmax.values())))))

